#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include <iostream.h>
#include <strstream.h>
#include <fstream.h>

#include "file.h"
#include "util.h"

// Parser Stuff

#undef yyFlexLexer
#define yyFlexLexer ffFlexLexer
#include <FlexLexer.h>

ffFlexLexer* fflexx = NULL;    // used by frlex
static FitsFile* ff = NULL;    // used by frerror
extern int ffparse(void*);

int fflex()
{
  return (fflexx ? fflexx->yylex() : 0);
}

void fferror(const char* m)
{
  if (ff) {
    ff->error(m);
    const char* cmd = fflexx ? fflexx->YYText() : (const char*)NULL;
    if (cmd && cmd[0] != '\n') {
      ff->error(": ");
      ff->error(cmd);
    }
  }
}

FitsFile::FitsFile()
{
  head_ = NULL;
  primary_ = NULL;
  managePrimary_ = 1;

  data_ = NULL;
  valid_ = 0;

  pName_ = NULL;
  pExt_ = NULL;
  pIndex_ = 0;
  pFilter_ = NULL;
  pBinX_ = NULL;
  pBinY_ = NULL;
  pBinZ_ = NULL;

  pWidth_ = 0;
  pHeight_ = 0;
  pBitpix_ = 0;
  pSkip_ = 0;
  pArch_ = NATIVE;

  byteswap_ = ftlsb();
}

FitsFile::~FitsFile()
{
  if (head_)
    delete head_;

  if (managePrimary_ && primary_)
    delete primary_;

  if (pName_)
    delete pName_;

  if (pExt_)
    delete pExt_;

  if (pFilter_)
    delete pFilter_;

  if (pBinX_)
    delete pBinX_;

  if (pBinY_)
    delete pBinY_;

  if (pBinZ_)
    delete pBinZ_;
}

void FitsFile::parse(const char* fn)
{
  if (fn) {
    istrstream str(fn);
    fflexx = new ffFlexLexer(&str);
    ff = this;
    ffparse(this);

    delete fflexx;
    fflexx = NULL;
  }

  if (!pBinX_ && !pBinY_) {
    char *env;
    if ((env = getenv("DS9_BINKEY"))) {
      istrstream str(env);
      fflexx = new ffFlexLexer(&str);
      ff = this;
      ffparse(this);

      delete fflexx;
      fflexx = NULL;
    }
  }

  if (!pWidth_ && !pHeight_ && !pBitpix_) {
    char *env;
    if ((env = getenv("DS9_ARRAY"))) {
      istrstream str(env);
      fflexx = new ffFlexLexer(&str);
      ff = this;
      ffparse(this);

      delete fflexx;
      fflexx = NULL;
    }
  }
}

void FitsFile::error(const char* m)
{
  // do nothing for now
  // cerr << m << endl;
}

int FitsFile::findEnd(const char* blk)
{
  for (int j=0; j<FTY_BLOCK; j+=FTY_CARDLEN)
    if (!strncmp("END      ", blk+j,8))
      return 1;

  return 0;
}

void FitsFile::setpName(const char* n)
{
  if (pName_)
    delete pName_;

  pName_ = ftdupstr(n);
}

void FitsFile::setpExt(const char* n)
{
  if (pExt_)
    delete pExt_;

  pExt_ = ftdupstr(n);
}

void FitsFile::setpFilter(const char* filt)
{
  if (pFilter_)
    delete pFilter_;

  pFilter_ = ftdupstr(filt);
}

void FitsFile::setpBinX(const char* x)
{
  if (pBinX_)
    delete pBinX_;

  pBinX_ = ftdupstr(x);
}

void FitsFile::setpBinY(const char* y)
{
  if (pBinY_)
    delete pBinY_;

  pBinY_ = ftdupstr(y);
}

void FitsFile::setpBinZ(const char* z)
{
  if (pBinZ_)
    delete pBinZ_;

  pBinZ_ = ftdupstr(z);
}

int FitsFile::validArrayParams()
{
  if (!pWidth_ || !pHeight_ || !pBitpix_)
    return 0;

  // check for valid bitpix
  switch (pBitpix_) {
  case 8:
  case 16:
  case -16:
  case 32:
  case -32:
  case -64:
    break;
  default:
    return 0;
  }

  return 1;
}

void FitsFile::setArrayByteSwap()
{
  switch (pArch_) {
  case NATIVE:
    byteswap_ = 0;
    break;
  case BIGENDIAN:
    byteswap_ = ftlsb();
    break;
  case LITTLEENDIAN:
    byteswap_ = !ftlsb();
    break;
  }
}

int FitsFile::find(const char* name)
{
  if (head_)
    if (head_->find(name))
      return 1;
    else
      if (primary_)
	if (primary_->find(name))
	  return 1;

  return 0;
}

int FitsFile::getLogical(const char* name, int def)
{
  if (head_) {
    int r = head_->getLogical(name,def);
    if (r != def)
      return r;
    else
      if (primary_)
	return primary_->getLogical(name,def);
  }

  return def;
}

int FitsFile::getInteger(const char* name, int def)
{
  if (head_) {
    int r = head_->getInteger(name,def);
    if (r != def)
      return r;
    else
      if (primary_)
	return primary_->getInteger(name,def);
  }

  return def;
}

double FitsFile::getReal(const char* name, double def)
{
  if (head_) {
    double r = head_->getReal(name,def);
    if (r != def)
      return r;
    else
      if (primary_)
	return primary_->getReal(name,def);
  }

  return def;
}

void FitsFile::getComplex(const char* name, double* real, double* img,
			  double rdef, double idef)
{
  if (head_) {
    head_->getComplex(name, real, img, rdef, idef);
    if (*real != rdef || *img != idef)
      return;
    else
      if (primary_) {
	primary_->getComplex(name, real, img, rdef, idef);
	return;
      }
  }

  *real = rdef;
  *img = idef;
}

char* FitsFile::getString(const char* name)
{
  if (head_) {
    char* r = head_->getString(name);
    if (r)
      return r;
    else
      if (primary_)
	return primary_->getString(name);
  }

  return NULL;
}

char* FitsFile::getComment(const char* name)
{
  if (head_) {
    char* r = head_->getComment(name);
    if (r)
      return r;
    else
      if (primary_)
	return primary_->getComment(name);
  }

  return NULL;
}

int FitsFile::saveFitsFile(const char* fn)
{
  if (!valid_)
    return 0;

  // we need this because windows will try to open as text
  ofstream fd(fn,ios::out|ios::binary);
  if (fd) {
    // write header
    // we have a problem here. the header may be an xtension, so lets force a
    // first line of 'SIMPLE' and then write the rest of the header
    char simple[80];
    memset(simple,' ',80);
    memcpy(simple,"SIMPLE  =                    T / Fits standard",46);
    fd.write(simple,80);
    fd.write(head()->cards()+80, head()->headbytes()-80);

    // write valid data
    // our data may be short (mmap or bad fits), so write valid data
    // then write the pad
    fd.write((char*)data(), head()->realbytes());

    // we may need to add a buffer to round out to block size
    int diff = head()->databytes() - head()->realbytes();
    if (diff>0) {
      char* buf = new char[diff];
      memset(buf,'\0',diff);
      fd.write(buf, diff);
      delete buf;
    }
    return 1;
  }
  else
    return 0;
}

int FitsFile::saveFitsChannel(Tcl_Interp* interp, const char* ch)
{
  if (!valid_)
    return 0;

  int tclMode;
  Tcl_Channel fd = Tcl_GetChannel(interp, (char*)ch, &tclMode);

  if (fd) {
    // write header
    // we have a problem here. the header may be an xtension, so lets force a
    // first line of 'SIMPLE' and then write the rest of the header
    char simple[80];
    memset(simple,' ',80);
    memcpy(simple,"SIMPLE  =                    T / Fits standard",46);
    Tcl_Write(fd, simple, 80);
    Tcl_Write(fd, head()->cards()+80, head()->headbytes()-80);

    // write valid data
    // our data may be short (mmap or bad fits), so write valid data
    // then write the pad
    Tcl_Write(fd, (char*)data(), head()->realbytes());

    // we may need to add a buffer to round out to block size
    int diff = head()->databytes() - head()->realbytes();
    if (diff>0) {
      char* buf = new char[diff];
      memset(buf,'\0',diff);
      Tcl_Write(fd, buf, diff);
      delete buf;
    }
    Tcl_Flush(fd);

    return 1;
  }
  else
    return 0;
}
