#include "ellipse.h"
#include "frame.h"
#include "util.h"

Ellipse::Ellipse(Frame* p, const Vector& ctr, const Vector& r, double a, 
		 const char* clr,  const char* f, const char* t, 
		 unsigned short prop)
  : BaseEllipse(p, ctr, a, clr, f, t, prop)
{
  annuli = 1;
  radii = new Vector[1];
  radii[0] = r;

  strcpy(type,"ellipse");
  handle = new Vector[4];
  numHandle = 4;

  updateBBox();
}

void Ellipse::edit(const Vector& v, int h)
{
  radii[0] = (v * Translate(-center) * flip * Rotate(-angle)).abs();

  updateBBox();
  doCallBacks(&editCB);
}

void Ellipse::setRadius(const Vector& r)
{
  radii[0] = r;
  updateBBox();

  doCallBacks(&editCB);
}

// private

void Ellipse::updateHandles()
{
  Vector r = radii[0];
  if (properties & FIXED)
    r /= parent->getZoom();

  Matrix m = Rotate(angle) * flip * Translate(center) * parent->refToCanvas;

  handle[0] = Vector(-r[0],-r[1]) * m;
  handle[1] = Vector( r[0],-r[1]) * m;
  handle[2] = Vector( r[0], r[1]) * m;
  handle[3] = Vector(-r[0], r[1]) * m;
}

// list

void Ellipse::list(ostream& str, int sys, int format, const char* delim)
{
  listCoordSystem(str, sys);
  str << ';';

  if (!(properties&INCLUDE))
    str << '-';

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector r = parent->mapLenFromRef(radii[0],(Frame::CoordSystem)sys);
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << r[0] << ',' << r[1] << ',' << radToDeg(angle) << ')';
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    Vector r = parent->mapLenFromRef(radii[0], (Frame::CoordSystem)sys,
				     Frame::ARCSEC);
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ','
	    << r[0] << "\"" << ',' << r[1] << "\"" << ',' 
	    << radToDeg(angle) << ')';
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << '(' << ra << ',' << dec << ',' 
	    << r[0] << "\"" << ',' << r[1] << "\"" << ',' 
	    << radToDeg(angle) << ')';
      }
      break;
    }
    break;
  }

  listProperties(str);
  str << delim;
}

void Ellipse::listCiao(ostream& str, int sys, int format, const char* delim)
{
  if (!(properties&INCLUDE))
    str << '-';

  // radius is always in image coords

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector r = parent->mapLenFromRef(radii[0], (Frame::CoordSystem)sys);
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << r[0] << ',' << r[1] << ',' << radToDeg(angle) << ')';
    }
    break;
  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    Vector r = parent->mapLenFromRef(radii[0], (Frame::CoordSystem)sys,
				     Frame::ARCSEC);
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << '(' << setprecision(8) << v[0] << "d," << v[1] << "d,"
	    << r[0] << "\"" << ',' << r[1] << "\"" << ',' 
	    << radToDeg(angle) << ')';
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << '(' << ra << ',' << dec << ',' 
	    << r[0] << "\"" << ',' << r[1] << "\"" << ',' 
	    << radToDeg(angle) << ')';
      }
      break;
    }
    break;
  }

  str << delim;
}

void Ellipse::listFtools(ostream& str, int sys, int format, const char* delim)
{
  if (!(properties&INCLUDE))
    str << '-';

  Vector r = parent->mapLenFromRef(radii[0], (Frame::CoordSystem)sys);
  Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
  str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
      << r[0] << ',' << r[1] << ',' << radToDeg(angle) << ')';

  str << delim;
}

void Ellipse::listPros(ostream& str, int sys, int format, const char* delim)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector r = parent->mapLenFromRef(radii[0],(Frame::CoordSystem)sys);
      Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
      str << type << ' ' << setprecision(8) << v << r << radToDeg(angle);
    }
    break;

  case Frame::PHYSICAL:
    {
      str << "physical; ";
      Vector r = parent->mapLenFromRef(radii[0],(Frame::CoordSystem)sys);
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys); 
      str << type << ' ' << setprecision(8) << v << r << radToDeg(angle);
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    listProsCoordSystem(str,sys);
    str << "; ";

    // display wcs radius in ARCSEC
    Vector r = parent->mapLenFromRef(radii[0],
				     (Frame::CoordSystem)sys,
				     Frame::ARCSEC);

    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << ' ' << setprecision(8) << v[0] << "d " << v[1] << "d " 
	    << r[0] << "\" " << r[1] << "\" " << radToDeg(angle);
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	if (dec[0]=='+')
	  str << type << ' ' << ra << ' ' << dec+1 << ' ' 
	      << r[0] << "\" " << r[1] << "\" " << radToDeg(angle);
	else
	  str << type << ' ' << ra << ' ' << dec << ' ' 
	      << r[0] << "\" " << r[1] << "\" " << radToDeg(angle);
      }
      break;
    }
    break;
  }

  str << delim;
}

void Ellipse::listSAOtng(ostream& str, int sys, int format, const char* delim)
{
  if (text[0])
    str << '#' << text << delim;

  if (properties&INCLUDE)
    str << '+';
  else
    str << '-';

  // radius is always in image coords

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
      Vector r = parent->mapLenFromRef(radii[0],Frame::IMAGE);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << r[0] << ',' << r[1] << ',' << radToDeg(angle) << ')';
    }
    break;
  case Frame::PHYSICAL:
    {
      Vector v = parent->map(center,Frame::REF,Frame::PHYSICAL);
      Vector r = parent->mapLenFromRef(radii[0],Frame::PHYSICAL);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << r[0] << ',' << r[1] << ',' << radToDeg(angle) << ')';
    }
    break;
  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	Vector r = parent->mapLenFromRef(radii[0], Frame::IMAGE);
	str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ','
	    << r[0] << ',' << r[1] << ',' << radToDeg(angle) << ')';
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	Vector r = parent->mapLenFromRef(radii[0], Frame::IMAGE);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << '(' << ra << ',' << dec << ',' 
	    << r[0] << ',' << r[1] << ',' << radToDeg(angle) << ')';
      }
      break;
    }
    break;
  }

  str << " # " << colorName << delim;
}

void Ellipse::listSAOimage(ostream& str, const char* delim)
{
  // all coords are in image coords

  if (!(properties&INCLUDE))
    str << '-';

  Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
  str << "ellipse(" << setprecision(8) << v[0] << ',' << v[1] 
      << ',' << radii[0][0] << ',' << radii[0][1] << ')';

  str << delim;
}
