#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <iostream.h>
#include <math.h>

#include "fitsdata.h"
#include "colorscale.h"
#include "nan.h"
#include "util.h"

// smallest permissible sample
// max frac. of pixels to be rejected
// k-sigma pixel rejection factor
// maximum number of fitline iterations

#define MIN_NPIXELS 5
#define MAX_REJECT 0.5
#define KREJ 2.5
#define MAX_ITERATIONS 5
#define INDEF 0

FitsData::FitsData(FitsFile* fits)
{
  FitsHead* head = fits->head();
  FitsImageHDU* hdu = (FitsImageHDU*)head->hdu();
  width = head->naxis(1);
  height = head->naxis(2);

  size = width * height;
  buf[0] = '\0';

  byteswap = fits->byteswap();

  bscale = hdu->bscale();
  bzero = hdu->bzero();
  hasScaling = hdu->hasscaling();
  hasBlank = hdu->blank();
  blank = hdu->blank();

  nanf = getnanf();
  nand = getnand();

  low = high = 0;

  zLow = zHigh = 0;
  uLow = uHigh = 0;

  scanValid = 0;

  incr = 100;
  sampleValid = 0;

  zContrast = .5;
  zSampleSize = 600;
  zSampleLine = 5;
  zscaleValid = 0;

  clipMode = Frame::MINMAX;
  mmMode = Frame::SAMPLE;

  if (fits->find("DATAMIN") && fits->find("DATAMAX")) {
    hasdatamin = 1;
    datamin = fits->getReal("DATAMIN", 0);
    datamax = fits->getReal("DATAMAX", 0);
  }
  else {
    hasdatamin = 0;
    datamin = datamax = 0;
  }

  if (fits->find("IRAF-MIN") && fits->find("IRAF-MAX")) {
    hasirafmin = 1;
    irafmin = fits->getReal("IRAF-MIN", 0);
    irafmax = fits->getReal("IRAF-MAX", 0);
  }
  else {
    hasirafmin = 0;
    irafmin = irafmax = 0;
  }
}

FitsData::~FitsData()
{}

const char* FitsData::getLow()
{
  ostrstream str(buf,32);
  str << low << ends;
  return buf;
}

const char* FitsData::getHigh()
{
  ostrstream str(buf,32);
  str << high << ends;
  return buf;
}

template<class T> FitsDatam<T>::FitsDatam(FitsFile* fits) : FitsData(fits)
{
  data = (T*)fits->data();
  min=max=0;
  mins=maxs=0;
  minX=minY=maxX=maxY=0;
}

template<class T> FitsDatam<T>::~FitsDatam()
{}

unsigned char FitsDatam<unsigned char>::swap(unsigned char* ptr)
{
  return *ptr;
}

short FitsDatam<short>::swap(short* ptr)
{
  short r;
  swap2((char*)ptr, (char*)&r);
  return r;
}

unsigned short FitsDatam<unsigned short>::swap(unsigned short* ptr)
{
  unsigned short r;
  swap2((char*)ptr, (char*)&r);
  return r;
}

BITPIX32 FitsDatam<BITPIX32>::swap(BITPIX32* ptr)
{
  BITPIX32 r;
  swap4((char*)ptr, (char*)&r);
  return r;
}

float FitsDatam<float>::swap(float* ptr)
{
  float r;
  swap4((char*)ptr, (char*)&r);
  return r;
}

double FitsDatam<double>::swap(double* ptr)
{
  double r;
  swap8((char*)ptr, (char*)&r);
  return r;
}

template<class T> void FitsDatam<T>::output(ostrstream& str, T value)
{
  str << value << ends;
}

void FitsDatam<unsigned char>::output(ostrstream& str, unsigned char value)
{
  str << (unsigned short)value << ends;
}

void FitsDatam<unsigned short>::output(ostrstream& str, unsigned short value)
{
  str << (unsigned short)value << ends;
}

template<class T> const char* FitsDatam<T>::getValue(const Vector& vv)
{
  Vector v(vv);

  int x = (int)v[0];
  int y = (int)v[1];

  ostrstream str(buf,32);
  if (x >= 0 && x < width && y >= 0 && y < height) {
    register T value = !byteswap ? data[y*width + x] : 
      swap(data+(y*width + x));

    if (hasBlank && value == blank) {
      str << "NaN" << ends;
      return buf;
    }

    if (hasScaling)
      str << value * bscale + bzero << ends;
    else
      output(str, value);
  }
  else
    str << ends;

  return buf;
}

template<class T> float FitsDatam<T>::getValueFloat(const Vector& v)
{
  Vector r = v;
  int x = (int)r[0];
  int y = (int)r[1];

  if (x >= 0 && x < width && y >= 0 && y < height) {
    register T value = !byteswap ? data[y*width + x] : 
      swap(data+(y*width + x));

    if (hasBlank && value == blank)
      return nand;

    return hasScaling ? value * bscale + bzero : value;
  }
  else
    return FLT_MIN;
}

template<class T> double FitsDatam<T>::getValueDouble(const Vector& v)
{
  Vector r = v;
  int x = (int)r[0];
  int y = (int)r[1];

  if (x >= 0 && x < width && y >= 0 && y < height) {
    register T value = !byteswap ? data[y*width + x] : 
      swap(data+(y*width + x));

    if (hasBlank && value == blank)
      return nand;

    return hasScaling ? value * bscale + bzero : value;
  }
  else
    return DBL_MIN;
}

template<class T> const char* FitsDatam<T>::getMin()
{
  // we do not check for blank/nan since this should never be set to nan

  ostrstream str(buf,32);

  switch (mmMode) {
  case Frame::SCAN:
    if (scanValid) {
      if (hasScaling)
	str << min * bscale + bzero << ends;
      else
	output(str, min);
    }
    else
      str << ends;
    break;
  case Frame::SAMPLE:
    if (sampleValid) {
      if (hasScaling)
	str << mins * bscale + bzero << ends;
      else
	output(str, mins);
    }
    else
      str << ends;
    break;
  case Frame::DATAMIN:
    if (hasdatamin)
      str << datamin << ends;
    else
      str << ends;
    break;
  case Frame::IRAFMIN:
    if (hasirafmin)
      str << irafmin << ends;
    else
      str << ends;
    break;
  }

  return buf;
}

template<class T> const char* FitsDatam<T>::getMax()
{
  // we do not check for blank/nan since this should never be set to nan

  ostrstream str(buf,32);

  switch (mmMode) {
  case Frame::SCAN:
    if (scanValid) {
      if (hasScaling)
	str << max * bscale + bzero << ends;
      else
	output(str, max);
    }
    else
      str << ends;
    break;
  case Frame::SAMPLE:
    if (sampleValid) {
      if (hasScaling)
	str << maxs * bscale + bzero << ends;
      else
	output(str, maxs);
    }
    else
      str << ends;
    break;
  case Frame::DATAMIN:
    if (hasdatamin)
      str << datamax << ends;
    else
      str << ends;
    break;
  case Frame::IRAFMIN:
    if (hasirafmin)
      str << irafmax << ends;
    else
      str << ends;
    break;
  }

  return buf;
}

template<class T> double FitsDatam<T>::getMinDouble()
{
  // we do not check for blank/nan since this should never be set to nan

  switch (mmMode) {
  case Frame::SCAN:
    if (scanValid) {
      if (hasScaling)
	return min * bscale + bzero;
      else
	return min;
    }
    else
      return 0;
  case Frame::SAMPLE:
    if (sampleValid) {
      if (hasScaling)
	return mins * bscale + bzero;
      else
	return mins;
    }
    else
      return 0;
  case Frame::DATAMIN:
    if (hasdatamin)
      return datamin;
    else
      return 0;
  case Frame::IRAFMIN:
    if (hasirafmin)
      return irafmin;
    else
      return 0;
  }
}

template<class T> double FitsDatam<T>::getMaxDouble()
{
  // we do not check for blank/nan since this should never be set to nan

  switch (mmMode) {
  case Frame::SCAN:
    if (scanValid) {
      if (hasScaling)
	return max * bscale + bzero;
      else
	return max;
    }
    else
      return 0;
  case Frame::SAMPLE:
    if (sampleValid) {
      if (hasScaling)
	return maxs * bscale + bzero;
      else
	return maxs;
    }
    else
      return 0;
  case Frame::DATAMIN:
    if (hasdatamin)
      return datamax;
    else
      return 0;
  case Frame::IRAFMIN:
    if (hasirafmin)
      return irafmax;
    return 0;
  }
}

template<class T> void FitsDatam<T>::updateMinMax(Frame::MinMaxMode mm, 
						  int mmi)
{
  mmMode = mm;

  if (incr != mmi)
    sampleValid = 0;

  incr = mmi;

  switch (mmMode) {
  case Frame::SCAN:
    if (!scanValid) {
      scan();
      scanValid = 1;
    }
    break;
  case Frame::SAMPLE:
    if (!sampleValid) {
      sample();
      sampleValid = 1;
    }
    break;
  }
}

template<class T> void FitsDatam<T>::updateClip(Frame::ClipMode cm, 
						Frame::MinMaxMode mm,
	float zc, int zs, int zl, int mmi, double ul, double uh, int fs)
{
  // we do not check for blank/nan since this should never be set to nan

  clipMode = cm;
  uLow = ul;
  uHigh = uh;

  // if force scan, go ahead and do it, else, set up minmax, just in case

  if (fs)
    updateMinMax(mm, mmi);
  else {
    mmMode = mm;
    if (incr != mmi)
      sampleValid = 0;
    incr = mmi;
  }

  if (zContrast != zc || zSampleSize != zs || zSampleLine != zl)
    zscaleValid = 0;

  zContrast = zc;
  zSampleSize = zs;
  zSampleLine = zl;


  switch (clipMode) {
  case Frame::MINMAX:
    switch (mmMode) {
    case Frame::SCAN:
      if (!scanValid) {
	scan();
	scanValid = 1;
      }
      if (hasScaling) {
	low = min * bscale + bzero;
	high = max * bscale + bzero;
      }
      else {
	low = min;
	high = max;
      }
      break;
    case Frame::SAMPLE:
      if (!sampleValid) {
	sample();
	sampleValid = 1;
      }
      if (hasScaling) {
	low = mins * bscale + bzero;
	high = maxs * bscale + bzero;
      }
      else {
	low = mins;
	high = maxs;
      }
      break;
    case Frame::DATAMIN:
      low = datamin;
      high = datamax;
      break;
    case Frame::IRAFMIN:
      low = irafmin;
      high = irafmax;
      break;
    }

    break;
  case Frame::ZSCALE:
    if (!zscaleValid) {
      if (DebugPerf)
	cerr << "zscale...";

      zscale();
      zscaleValid = 1;

      if (DebugPerf)
	cerr << "end" << endl;
    }
    low = zLow;
    high = zHigh;

    break;
  case Frame::USERCLIP:
    low = uLow;
    high = uHigh;

    break;
  }
}

template<class T> void FitsDatam<T>::bin(double* arr, int length, 
					 double mn, double mx)
{
  T* ptr = data;
  double diff = mx-mn;
  int last = length-1;

  int offset = 1;
  if (mmMode == Frame::SAMPLE)
    offset = incr;

  for (long i=0; i<size; i+=offset, ptr++) {
    register double value = !byteswap ? *ptr : swap(ptr);
    if (hasBlank && value == blank)
      continue; // skip nan's

    if (hasScaling)
      value = value * bscale + bzero;

    if (value < mn)
      arr[0]++;
    else if (value > mx)
      arr[last]++;
    else
      arr[(int)(((value-mn)/diff)*last+.5)]++;
  }
}

template<class T> void FitsDatam<T>::scan()
{
  if (DebugPerf)
    cerr << "scan...";

  min = Tmax();
  max = Tmin();

  T* ptr = data;
  for (int j=0; j<height; j++)
    for (int i=0; i<width; i++, ptr++) {
      register T value = !byteswap ? *ptr : swap(ptr);

      if (hasBlank && value == blank)
	continue; // skip nan's 

      if (value < min) {
	min = value;
	minX=i;
	minY=j;
      }
      else if (value > max) {
	max = value;
	maxX = i;
	maxY = j;
      }
    }

  if (DebugPerf) {
    cerr << "end" << endl;
    cerr << "min(" << minX << "," << minY << ")=" << min << endl;
    cerr << "max(" << maxX << "," << maxY << ")=" << max << endl;
  }
}

template<class T> void FitsDatam<T>::sample()
{
  if (DebugPerf)
    cerr << "sample...";

  mins = Tmax();
  maxs = Tmin();

  T* ptr = data;
  for (int a=0; a<size; a+=incr, ptr+=incr) {
    register T value = !byteswap ? *ptr : swap(ptr);

    if (hasBlank && value == blank)
      continue; // skip nan's 

    if (value < mins) {
      mins = value;
      minX = (int)(fmod(a,width));
      minY = a/width;
    }
    else if (value > maxs) {
      maxs = value;
      maxX = (int)(fmod(a,width));
      maxY = a/width;
    }
  }

  if (DebugPerf) {
    cerr << "end" << endl;
    cerr << "mins(" << minX << "," << minY << ")=" << mins << endl;
    cerr << "maxs(" << maxX << "," << maxY << ")=" << maxs << endl;
  }
}

unsigned char FitsDatam<unsigned char>::Tmin() {return 0;}
unsigned char FitsDatam<unsigned char>::Tmax() {return UCHAR_MAX;}

short FitsDatam<short>::Tmin() {return SHRT_MIN;}
short FitsDatam<short>::Tmax() {return SHRT_MAX;}

unsigned short FitsDatam<unsigned short>::Tmin() {return 0;}
unsigned short FitsDatam<unsigned short>::Tmax() {return USHRT_MAX;}

#if defined( __alpha) || defined(__mips64)
int FitsDatam<int>::Tmin() {return INT_MIN;}
int FitsDatam<int>::Tmax() {return INT_MAX;}
#else
long FitsDatam<long>::Tmin() {return LONG_MIN;}
long FitsDatam<long>::Tmax() {return LONG_MAX;}
#endif

float FitsDatam<float>::Tmin() {return FLT_MIN;}
float FitsDatam<float>::Tmax() {return FLT_MAX;}

double FitsDatam<double>::Tmin() {return DBL_MIN;}
double FitsDatam<double>::Tmax() {return DBL_MAX;}

// ZSCALE

#undef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#undef min
#define min(a,b) ((a) < (b) ? (a) : (b)) 
#undef mod
#define mod(a,b) ((a) % (b))
#undef nint
#define nint(a) ((int)(a + 0.5))

// ZSCALE -- Compute the optimal Z1, Z2 (range of greyscale values to be
// displayed) of an image.  For efficiency a statistical subsample of an image
// is used.  The pixel sample evenly subsamples the image in x and y.  The
// entire image is used if the number of pixels in the image is smaller than
// the desired sample.
//
// The sample is accumulated in a buffer and sorted by greyscale value.
// The median value is the central value of the sorted array.  The slope of a
// straight line fitted to the sorted sample is a measure of the standard
// deviation of the sample about the median value.  Our algorithm is to sort
// the sample and perform an iterative fit of a straight line to the sample,
// using pixel rejection to omit gross deviants near the endpoints.  The fitted
// straight line is the transfer function used to map image Z into display Z.
// If more than half the pixels are rejected the full range is used.  The slope
// of the fitted line is divided by the user-supplied contrast factor and the
// final Z1 and Z2 are computed, taking the origin of the fitted line at the
// median value.

template<class T> void FitsDatam<T>::zscale()
{
  // Subsample the image

  float* sample;
  int npix = zSampleImage(&sample);
  int center_pixel = max(1, (npix + 1) / 2);

  // Sort the sample, compute the minimum, maximum, and median pixel values

  qsort((void*)sample, npix, sizeof(float), fCompare);
  float zmin = *sample;
  float zmax = *(sample+max(npix,1)-1);

  // The median value is the average of the two central values if there 
  // are an even number of pixels in the sample.

  float* left = &(sample[center_pixel - 1]);

  float median;
  if (mod(npix, 2) == 1 || center_pixel >= npix)
    median = *left;
  else
    median = (*left + *(left+1)) / 2;

  // Fit a line to the sorted sample vector.  If more than half of the
  // pixels in the sample are rejected give up and return the full range.
  // If the user-supplied contrast factor is not 1.0 adjust the scale
  // accordingly and compute zLow and zHigh, the y intercepts at indices 1 and
  // npix.

  int minpix = max(MIN_NPIXELS, (int)(npix * MAX_REJECT));
  int ngrow = max(1, nint(npix * .01));
  float zstart, zslope;

  int ngoodpix = zFitLine(sample, npix, &zstart, &zslope, 
			  KREJ, ngrow, MAX_ITERATIONS);

  if (ngoodpix < minpix) {
    zLow = zmin;
    zHigh = zmax;
  }
  else {
    if (zContrast > 0)
      zslope = zslope / zContrast;
    zLow = max(zmin, median - (center_pixel - 1) * zslope);
    zHigh = min(zmax, median + (npix - center_pixel) * zslope);
  }

  delete [] sample;
}

// sampleImage -- Extract an evenly gridded subsample of the pixels from
// a two-dimensional image into a one-dimensional vector.

template<class T> int FitsDatam<T>::zSampleImage(float** sample)
{
  // Compute the number of pixels each line will contribute to the sample,
  // and the subsampling step size for a line.  The sampling grid must
  // span the whole line on a uniform grid.

  int opt_npix_per_line = max(1, min(width, zSampleLine));
  int col_step = max(2, (width + opt_npix_per_line-1) / opt_npix_per_line);
  int npix_per_line = max(1, (width + col_step-1) / col_step);

  // Compute the number of lines to sample and the spacing between lines.
  // We must ensure that the image is adequately sampled despite its
  // size, hence there is a lower limit on the number of lines in the
  // sample.  We also want to minimize the number of lines accessed when
  // accessing a large image, because each disk seek and read is ex-
  // pensive. The number of lines extracted will be roughly the sample
  // size divided by zSampleLine, possibly more if the lines are very
  // short.

  int min_nlines_in_sample = max(1, zSampleSize / zSampleLine);
  int opt_nlines_in_sample = max(min_nlines_in_sample, 
				 min(height, (zSampleSize + npix_per_line-1) / 
				 npix_per_line));
  int line_step = max(2, height / opt_nlines_in_sample);
  int max_nlines_in_sample = (height + line_step-1) / line_step;

  // Allocate space for the output vector.  Buffer must be freed by our caller.

  int maxpix = npix_per_line * max_nlines_in_sample;
  *sample = new float[maxpix];
  float* row = new float[width];

  // Extract the vector

  int npix = 0;
  float* op = *sample;

  for (int line = (line_step + 1)/2; line < height; line+=line_step) {

    // Load a row of values from the image

    for (int i=0; i < width; i++) {
      T value = !byteswap ? data[(line-1)*width + i] : 
	swap(data+(line-1)*width + i);

      if (hasBlank && (value == blank)) {
	row[i] = nanf;
	continue;
      }	

      row[i] = hasScaling ? value * bscale + bzero : value;
    }    

    int got = zSubSample(row, op, npix_per_line, col_step);
    op += got;
    npix += got;
    if (npix >= maxpix)
      break;
  }

  delete [] row;
  return npix;
}

// subSample -- Subsample an image line.  Extract the first pixel and
// every "step"th pixel thereafter for a total of npix pixels.

int FitsData::zSubSample(float* a, float* b, int npix, int step)
{
  if (step <= 1)
    step = 1;

  int got = 0;
  int ip = 0;
  for (int i=0; i<npix; i++) {
    if (!isnanf(a[ip])) // we skip over the NaN pixels
      b[got++] = a[ip];

    ip += step;
  }

  return got;
}

// fitLine -- Fit a straight line to a data array of type real.  This is
// an iterative fitting algorithm, wherein points further than ksigma from the
// current fit are excluded from the next fit.  Convergence occurs when the
// next iteration does not decrease the number of pixels in the fit, or when
// there are no pixels left.  The number of pixels left after pixel rejection
// is returned as the function value.

int FitsData::zFitLine (float* sampleData, int npix, float* zstart, 
			float* zslope, float krej, int ngrow, int maxiter)
{
  float	xscale;
  if (npix <= 0)
    return (0);
  else if (npix == 1) {
    *zstart = sampleData[1];
    *zslope = 0.0;

    return (1);
  }
  else
    xscale = 2.0 / (npix - 1);

  // Allocate a buffer for data minus fitted curve, another for the
  // normalized X values, and another to flag rejected pixels.

  float* flat = new float[npix];
  float* normx = new float[npix];
  short* badpix = new short[npix];

  for (int k=0; k<npix; k++)
    badpix[k]=0;

  // Compute normalized X vector.  The data X values [1:npix] are
  // normalized to the range [-1:1].  This diagonalizes the lsq matrix
  // and reduces its condition number.

  for (int i=0; i<npix; i++)
    normx[i] = i * xscale - 1.0;

  // Fit a line with no pixel rejection.  Accumulate the elements of the
  // matrix and data vector.  The matrix M is diagonal with
  // M[1,1] = sum x**2 and M[2,2] = ngoodpix.  The data vector is
  // DV[1] = sum (data[i] * x[i]) and DV[2] = sum (data[i]).

  double sumxsqr = 0;
  double sumxz = 0;
  double sumx = 0;
  double sumz = 0;

  for (int j=0; j<npix; j++) {
    float x = normx[j];
    float z = sampleData[j];
    sumxsqr = sumxsqr + (x * x);
    sumxz   = sumxz + z * x;
    sumz    = sumz + z;
  }

  // Solve for the coefficients of the fitted line

  float z0 = sumz / npix;
  float dz = sumxz / sumxsqr;

  // Iterate, fitting a new line in each iteration. Compute the flattened
  // data vector and the sigma of the flat vector.  Compute the lower and
  // upper k-sigma pixel rejection thresholds.  Run down the flat array
  // and detect pixels to be rejected from the fit.  Reject pixels from
  // the fit by subtracting their contributions from the matrix sums and
  // marking the pixel as rejected.

  int ngoodpix = npix;
  int last_ngoodpix;
  int minpix = max (MIN_NPIXELS, (int) (npix * MAX_REJECT));

  for (int niter=0;  niter < maxiter;  niter++) {
    last_ngoodpix = ngoodpix;

    // Subtract the fitted line from the data array

    zFlattenData(sampleData, flat, normx, npix, z0, dz);

    // Compute the k-sigma rejection threshold.  In principle this
    // could be more efficiently computed using the matrix sums
    // accumulated when the line was fitted, but there are problems with
    // numerical stability with that approach.

    float mean;
    float sigma;
    ngoodpix = zComputeSigma (flat, badpix, npix, &mean, &sigma);
    float threshold = sigma * krej;

    // Detect and reject pixels further than ksigma from the fitted
    // line.

    ngoodpix = zRejectPixels(sampleData, flat, normx,
			     badpix, npix, &sumxsqr, &sumxz, &sumx, &sumz, 
			     threshold, ngrow);

    // Solve for the coefficients of the fitted line.  Note that after
    // pixel rejection the sum of the X values need no longer be zero.

    if (ngoodpix > 0) {
      double rowrat = sumx / sumxsqr;
      z0 = (sumz - rowrat * sumxz) / (ngoodpix - rowrat * sumx);
      dz = (sumxz - z0 * sumx) / sumxsqr;
    }

    if (ngoodpix >= last_ngoodpix || ngoodpix < minpix)
      break;
  }

  // Transform the line coefficients back to the X range [1:npix]

  *zstart = z0 - dz;
  *zslope = dz * xscale;

  delete [] flat;
  delete [] normx;
  delete [] badpix;

  return ngoodpix;
}


// flattenData -- Compute and subtract the fitted line from the data array,
// returned the flattened data in FLAT.

void FitsData::zFlattenData(float* sampleData, float* flat, float* x, 
			     int npix, float z0, float dz)
{
  for (int i=0; i < npix; i++) 
    flat[i] = sampleData[i] - (x[i] * dz + z0);
}

// computeSigma -- Compute the root mean square deviation from the
// mean of a flattened array.  Ignore rejected pixels.

int FitsData::zComputeSigma(float* a, short* badpix, int npix, 
			    float* mean, float* sigma)
{
  int ngoodpix = 0;
  double sum = 0.0;
  double sumsq = 0.0;

  // Accumulate sum and sum of squares

  for (int i=0; i < npix; i++)
    if (badpix[i] == GOOD_PIXEL) {
      float pixval = a[i];
      ngoodpix = ngoodpix + 1;
      sum = sum + pixval;
      sumsq = sumsq + pixval * pixval;
    }

  // Compute mean and sigma

  switch (ngoodpix) {
  case 0:
    *mean = INDEF;
    *sigma = INDEF;
    break;
  case 1:
    *mean = sum;
    *sigma = INDEF;
    break;
  default:
    *mean = sum / ngoodpix;
    double temp = sumsq / (ngoodpix-1) - (sum*sum) / (ngoodpix*(ngoodpix - 1));
    if (temp < 0)		// possible with roundoff error
      *sigma = 0.0;
    else
      *sigma = sqrt(temp);
  }

  return ngoodpix;
}

// rejectPixels -- Detect and reject pixels more than "threshold" greyscale
// units from the fitted line.  The residuals about the fitted line are given
// by the "flat" array, while the raw data is in "data".  Each time a pixel
// is rejected subtract its contributions from the matrix sums and flag the
// pixel as rejected.  When a pixel is rejected reject its neighbors out to
// a specified radius as well.  This speeds up convergence considerably and
// produces a more stringent rejection criteria which takes advantage of the
// fact that bad pixels tend to be clumped.  The number of pixels left in the
// fit is returned as the function value.

int FitsData::zRejectPixels(float* sampleData, float* flat, float *normx, 
			    short *badpix, int npix, double* sumxsqr, 
			    double* sumxz, double* sumx, double* sumz, 
			    float threshold, int ngrow)
{
  int ngoodpix = npix;
  float lcut = -threshold;
  float hcut = threshold;

  for (int i=0; i < npix; i++) {
    if (badpix[i] == BAD_PIXEL)
      ngoodpix = ngoodpix - 1;
    else {
      float residual = flat[i];
      if (residual < lcut || residual > hcut) {

	// Reject the pixel and its neighbors out to the growing
	// radius.  We must be careful how we do this to avoid
	// directional effects.  Do not turn off thresholding on
	// pixels in the forward direction; mark them for rejection
	// but do not reject until they have been thresholded.
	// If this is not done growing will not be symmetric.

	for (int j=max(0,i-ngrow); j < min(npix,i+ngrow); j++) {
	  if (badpix[j] != BAD_PIXEL) {
	    if (j <= i) {
	      double x = normx[j];
	      double z = sampleData[j];
	      *sumxsqr = *sumxsqr - (x * x);
	      *sumxz = *sumxz - z * x;
	      *sumx = *sumx - x;
	      *sumz = *sumz - z;
	      badpix[j] = BAD_PIXEL;
	      ngoodpix = ngoodpix - 1;
	    } else
	      badpix[j] = REJECT_PIXEL;
	  }
	}
      }
    }
  }

  return ngoodpix;
}

template class FitsDatam<unsigned char>;
template class FitsDatam<short>;
template class FitsDatam<unsigned short>;
template class FitsDatam<BITPIX32>;
template class FitsDatam<float>;
template class FitsDatam<double>;
